// Copyright 2017 Fortanix, Inc. All rights reserved.
//
// This example demonstrates how to connect to the SDKMS server as an
// application using an API key and perform encryption and decryption
// using batch encryption calls. The batch crypto APIs allow multiple
// different pieces of data to be encrypted or decrypted in a single
// call.

package com.fortanix.sdkms.v1.examples;

import java.io.IOException;
import java.net.HttpURLConnection;
import java.util.Base64;
import java.util.List;
import org.apache.commons.cli.*;
import org.apache.commons.io.IOUtils;

// API library imports.
import com.fortanix.sdkms.v1.ApiClient;
import com.fortanix.sdkms.v1.ApiException;
import com.fortanix.sdkms.v1.Configuration;
import com.fortanix.sdkms.v1.api.AuthenticationApi;
import com.fortanix.sdkms.v1.api.EncryptionAndDecryptionApi;
import com.fortanix.sdkms.v1.api.SecurityObjectsApi;
import com.fortanix.sdkms.v1.auth.ApiKeyAuth;
import com.fortanix.sdkms.v1.model.AuthResponse;
import com.fortanix.sdkms.v1.model.BatchDecryptRequest;
import com.fortanix.sdkms.v1.model.BatchDecryptRequestInner;
import com.fortanix.sdkms.v1.model.BatchDecryptResponse;
import com.fortanix.sdkms.v1.model.BatchEncryptRequest;
import com.fortanix.sdkms.v1.model.BatchEncryptRequestInner;
import com.fortanix.sdkms.v1.model.BatchEncryptResponse;
import com.fortanix.sdkms.v1.model.CryptMode;
import com.fortanix.sdkms.v1.model.DecryptRequest;
import com.fortanix.sdkms.v1.model.DecryptResponse;
import com.fortanix.sdkms.v1.model.EncryptRequest;
import com.fortanix.sdkms.v1.model.EncryptResponse;
import com.fortanix.sdkms.v1.model.KeyObject;
import com.fortanix.sdkms.v1.model.ObjectType;

// For convenience, you might want to just use the following imports instead:
//import com.fortanix.sdkms.v1.*;
//import com.fortanix.sdkms.v1.api.*;
//import com.fortanix.sdkms.v1.auth.*;
//import com.fortanix.sdkms.v1.models.*;

public class BatchCrypto {
    private boolean debug;
    private boolean debug_tls;
    private String server;
    private String truststore;
    private String apiKey;
    private String key1Name;
    private String key2Name;
    private String bearerToken;

    public void connect() {
        ApiClient client = new ApiClient();
        // Set the name of the server to talk to.
        client.setBasePath(server);

        // This optionally enables verbose logging in the API library.
        client.setDebugging(debug);

        // The default ApiClient (and its configured authorization) will be
        // used for constructing the specific API objects, such as
        // EncryptionAndDecryptionApi and SecurityObjectsApi.
        Configuration.setDefaultApiClient(client);

        // If you need a trust store for the server's certificate, you can
        // configure it here or via -Djavax.net.ssl.trustStore= on the
        // java command line.
        if (truststore != null) {
            System.setProperty("javax.net.ssl.trustStore", truststore);
        }

        // This optionally enables very verbose logging in the Java network
        // libraries.
        if (debug_tls) {
            System.setProperty("javax.net.debug", "all");
        }

        // When authenticating as an application, the API Key functions as
        // the entire HTTP basic auth token.
        client.setBasicAuthString(apiKey);

        // Acquire a bearer token to use for other APIs.
        try {
            AuthResponse response = new AuthenticationApi().authorize();
            bearerToken = response.getAccessToken();
            if (debug) {
                System.err.printf("Received Bearer token %s\n", bearerToken);
            }

            // Configure the client library to use the bearer token.
            ApiKeyAuth bearerAuth =
                (ApiKeyAuth) client.getAuthentication("bearerToken");
            bearerAuth.setApiKey(bearerToken);
            bearerAuth.setApiKeyPrefix("Bearer");
        } catch (ApiException e) {
            System.err.println("Unable to authenticate: " + e.getMessage());
            System.exit(1);
        }
    }

    public KeyObject findKey(String name) throws ApiException {
        // This filters by key name. Since names are unique, this will
        // return either 0 or 1 object.
        List<KeyObject> keys =
            new SecurityObjectsApi().getSecurityObjects(name, null, null);
        if (keys.size() == 0) {
            throw new ApiException("No key found with name " + name);
        } else {
            return keys.get(0);
        }
    }

    public void performCrypto() throws ApiException, IOException {
        String plain1 = "Hello, World!";
        String plain2 = "Hello, World Again!";

        System.out.println("Plaintext1: " + plain1);
        System.out.println("plaintext2: " + plain2);

        // Find a key by name or ID. You could skip this step if you already
        // know the ID of the key you wish to encrypt or decrypt with and
        // the key type.
        KeyObject key1 = findKey(key1Name);
        KeyObject key2 = findKey(key2Name);
        if (key1.getObjType() != ObjectType.AES ||
            key2.getObjType() != ObjectType.AES) {
            throw new ApiException("This example requires 1 or 2 AES keys.");
        }

        // Construct a batch request. Batches may contain any number of
        // requests, but the entire batch is subject to the request size
        // limit.
        BatchEncryptRequest encryptBatch = new BatchEncryptRequest();

        // Each piece of data to be encrypted gets its own EncryptRequest.
        // Keys, algorithms, and modes are all specified per-request, so a
        // single batch may use multiple different keys or even algorithms.
        EncryptRequest encryptRequest1 = new EncryptRequest()
            .plain(plain1.getBytes())
            .alg(ObjectType.AES)
            .mode(CryptMode.CBC);

        // Each request gets wrapped in a BatchEncryptRequestInner with the key.
        BatchEncryptRequestInner encryptInner1 = new BatchEncryptRequestInner()
            .kid(key1.getKid())
            .request(encryptRequest1);
        // Then the request is added to the batch.
        encryptBatch.add(encryptInner1);

        // Here we do the same thing with a possibly different key, using a
        // different cipher mode.
        EncryptRequest encryptRequest2 = new EncryptRequest()
            .plain(plain2.getBytes())
            .alg(ObjectType.AES)
            .mode(CryptMode.GCM)
            .tagLen(128);

        BatchEncryptRequestInner encryptInner2 = new BatchEncryptRequestInner()
            .kid(key2.getKid())
            .request(encryptRequest2);
        encryptBatch.add(encryptInner2);

        EncryptionAndDecryptionApi api = new EncryptionAndDecryptionApi();
        BatchEncryptResponse encryptResponse = api.batchEncrypt(encryptBatch);

        // The reponse is a list of EncryptResponseInner objects in the same
        // order as the requests in the BatchEncryptRequest. Note that it's
        // possible for some requests in a batch to succeed while others fail,
        // so you should check the status of each request separately. status
        // here is the HTTP status code for each subrequest.
        if (encryptResponse.get(0).getStatus() != HttpURLConnection.HTTP_OK ||
            encryptResponse.get(1).getStatus() != HttpURLConnection.HTTP_OK) {
            throw new ApiException("One of the encryption requests failed");
        }

        // The returned ciphertext is in binary, and probably not
        // human readable. We convert to base64 here to get something printable.
        System.out.println("Ciphertext1: " + Base64.getEncoder().encodeToString(
                               encryptResponse.get(0).getBody().getCipher()));
        System.out.println("Ciphertext2: " + Base64.getEncoder().encodeToString(
                               encryptResponse.get(1).getBody().getCipher()));

        // Now we will decrypt our ciphertexts using a batch decryption
        // operation. Batch decryption is similar to encryption, but uses
        // decrypt requests and responses instead of encrypt.
        BatchDecryptRequest decryptBatch = new BatchDecryptRequest();
        DecryptRequest decryptRequest1 = new DecryptRequest()
            .cipher(encryptResponse.get(0).getBody().getCipher())
            .alg(ObjectType.AES)
            .mode(CryptMode.CBC)
            .iv(encryptResponse.get(0).getBody().getIv());
        BatchDecryptRequestInner decryptInner1 = new BatchDecryptRequestInner()
            .kid(key1.getKid())
            .request(decryptRequest1);
        decryptBatch.add(decryptInner1);

        DecryptRequest decryptRequest2 = new DecryptRequest()
            .cipher(encryptResponse.get(1).getBody().getCipher())
            .alg(ObjectType.AES)
            .mode(CryptMode.GCM)
            .iv(encryptResponse.get(1).getBody().getIv())
            .tag(encryptResponse.get(1).getBody().getTag());
        BatchDecryptRequestInner decryptInner2 = new BatchDecryptRequestInner()
            .kid(key2.getKid())
            .request(decryptRequest2);
        decryptBatch.add(decryptInner2);

        BatchDecryptResponse decryptResponse = api.batchDecrypt(decryptBatch);
        // Like encryption, the decryption response contains a list of
        // BatchDecryptResponseInner objects, in the same order as the request.
        // Each response has its own status.
        if (decryptResponse.get(0).getStatus() != HttpURLConnection.HTTP_OK ||
            decryptResponse.get(1).getStatus() != HttpURLConnection.HTTP_OK) {
            throw new ApiException("One of the decryption requests failed");
        }

        String decrypted1 = new String(decryptResponse.get(0).getBody().
                                       getPlain(), "UTF-8");
        System.out.println("Decrypted plaintext1: " + decrypted1);
        String decrypted2 = new String(decryptResponse.get(1).getBody().
                                       getPlain(), "UTF-8");
        System.out.println("Decrypted plaintext2: " + decrypted2);
    }

    public void logout() {
        if (bearerToken != null) {
            // It is a good idea to terminate the session when you are done
            // using it. This minimizes the window of time in which an attacker
            // could steal your bearer token and use it.
            try {
                new AuthenticationApi().terminate();
            } catch (ApiException e) {
                System.err.println("Error logging out: " + e.getMessage());
            }
            bearerToken = null;
        }
    }

    protected void usage(Options options) {
        HelpFormatter formatter = new HelpFormatter();
        formatter.printHelp("BatchCrypto",
                            "This program demonstrates batch cryptography " +
                            "using one or two symmetric keys. The names of " +
                            "two AES keys must be passed as --key1-name and " +
                            "--key2-name. These may be the same key name. " +
                            "The program will encrypt and decrypt some " +
                            "sample data.", options, "");
    }

    protected void help(String msg) {
        System.err.println(msg);
        System.err.println("Run BatchCrypto --help for usage information");
    }

    public void parseCmdline(String[] args) {
        Options options = new Options();
        options.addOption(Option.builder("s")
                          .longOpt("server")
                          .hasArg()
                          .type(String.class)
                          .desc("SDKMS server endpoint (e.g. " +
                                "https://apps.smartkey.io)")
                          .build());

        options.addOption(Option.builder("k")
                          .longOpt("api-key")
                          .required(true)
                          .hasArg()
                          .type(String.class)
                          .desc("API key for this application")
                          .build());

        options.addOption(Option.builder("K")
                          .longOpt("key-id")
                          .hasArg()
                          .type(String.class)
                          .desc("Key ID of the cryptographic key to use")
                          .build());

        options.addOption(Option.builder()
                          .longOpt("key1-name")
                          .hasArg()
                          .required(true)
                          .type(String.class)
                          .desc("Name of first symmetric key to use")
                          .build());

        options.addOption(Option.builder()
                          .longOpt("key2-name")
                          .hasArg()
                          .required(true)
                          .type(String.class)
                          .desc("Name of second symmetric key to use")
                          .build());

        options.addOption(Option.builder("t")
                          .longOpt("truststore")
                          .hasArg()
                          .type(String.class)
                          .desc("Java truststore containing server certificate")
                          .build());

        options.addOption(Option.builder("d")
                          .longOpt("debug")
                          .desc("Enable client API library debugging")
                          .build());

        options.addOption(Option.builder("D")
                          .longOpt("debug-tls")
                          .desc("Enable Java TLS debugging")
                          .build());

        options.addOption(Option.builder("h")
                          .longOpt("help")
                          .desc("Get usage information and exit")
                          .build());

        CommandLineParser parser = new DefaultParser();
        try {
            CommandLine cmdline = parser.parse(options, args);
            if (cmdline.hasOption("help")) {
                usage(options);
                System.exit(0);
            }

            debug = cmdline.hasOption("debug");
            debug_tls = cmdline.hasOption("debug-tls");
            apiKey = cmdline.getOptionValue("api-key");

            if (cmdline.hasOption("server")) {
                server = cmdline.getOptionValue("server");
            } else {
                server = "https://apps.smartkey.io";
            }

            if (cmdline.hasOption("truststore")) {
                truststore = cmdline.getOptionValue("truststore");
            }

            key1Name = cmdline.getOptionValue("key1-name");
            key2Name = cmdline.getOptionValue("key2-name");
        } catch (ParseException e) {
            usage(options);
            System.err.println("Command-line parsing failed: " +
                               e.getMessage());
            System.exit(1);
        }
    }

    public static void main(String[] args) throws ApiException {
        int status = 0;

        BatchCrypto crypto = new BatchCrypto();
        crypto.parseCmdline(args);
        crypto.connect();
        try {
            crypto.performCrypto();
        } catch (ApiException | IOException e) {
            System.err.println("Crypto failed: " + e.getMessage());
            status = 1;
        } finally {
            crypto.logout();
        }
        System.exit(status);
    }
}
